#include "GBADecompress.h"
#include "agb.h"

#define getbit_8(bb, src, ilen) \
	(((bb = bb & 0x7f ? bb*2 : ((unsigned)src[ilen++]*2+1)) >> 8) & 1)

#define getbit(bb)	  getbit_8(bb,src,ilen)

#    define u32_C(c)     c ## U

int GBADecompress( const u8 * src, u32  src_len,u8 * dst, u32 * dst_len,void * wrkmem )
{
	u32 bb;
	u32 ilen, olen, last_m_off;

	bb = 0;
	ilen = 0, olen = 0, last_m_off = 1;

	for (;;)
	{
		u32 m_off, m_len;

		while (getbit(bb))
		{
			dst[olen++] = src[ilen++];
		}
		m_off = 1;
		do
		{
			m_off = m_off*2 + getbit(bb);
		} while (!getbit(bb));
		if (m_off == 2)
		{
			m_off = last_m_off;
		}
		else
		{
			m_off = (m_off-3)*256 + src[ilen++];
			if (m_off == u32_C(0xffffffff))
			{
				break;
			}
			last_m_off = ++m_off;
		}
		m_len = getbit(bb);
		m_len = m_len*2 + getbit(bb);
		if (m_len == 0)
		{
			m_len++;
			do
			{
				m_len = m_len*2 + getbit(bb);
			} while (!getbit(bb));
			m_len += 2;
		}
		m_len += (m_off > 0xd00);

		{
			const u8 *m_pos;
			m_pos = dst + olen - m_off;
			dst[olen++] = *m_pos++;
			do
			{
				dst[olen++] = *m_pos++;
			} while (--m_len > 0);
		}
	}
	*dst_len = olen;
	return ilen == src_len ? (0/*E_OK*/) : (-1);
}

static void InitCache(u8 *cache,int *cacheoffset,u8 *realmem,int cachesize)
{
	DmaWait(3);
	DmaCopy(3,realmem+(*cacheoffset),cache,cachesize,16);
	DmaWait(3);
}

static void FlushCache(u8 *cache,int *cacheoffset,u8 *realmem,int cachesize)
{
	DmaWait(3);
	DmaCopy(3,cache,realmem+(*cacheoffset),cachesize,16);
	DmaWait(3);
}

static void WriteOffset(u8 towrite,int writeoffset,u8 *cache,int *cacheoffset,u8 *realmem,int cachesize)
{
	if (writeoffset < (*cacheoffset) || writeoffset >= ((*cacheoffset)+cachesize) )
	{
		FlushCache(cache,cacheoffset,realmem,cachesize);
		*cacheoffset = writeoffset/cachesize;
		*cacheoffset *= cachesize;
		InitCache(cache,cacheoffset,realmem,cachesize);
	}
	cache[writeoffset - (*cacheoffset)] = towrite;
}

static u8 ReadOffset(int readoffset,u8 *cache,int *cacheoffset,u8 *realmem,int cachesize)
{
	if (readoffset < (*cacheoffset) || readoffset >= ((*cacheoffset)+cachesize) )
	{
		FlushCache(cache,cacheoffset,realmem,cachesize);
		*cacheoffset = readoffset/cachesize;
		*cacheoffset *= cachesize;
		InitCache(cache,cacheoffset,realmem,cachesize);
	}

	return cache[readoffset - (*cacheoffset)];
}

int GBADecompressVRAM( const u8 * src, u32  src_len,u8 * dst, u32 * dst_len,void * wrkmem )
{
	u32 bb;
	u32 ilen, olen, last_m_off;
	u32 writecache[4];
	int writeoffset;

	writeoffset = 0;
	InitCache((u8*)writecache,&writeoffset,dst,sizeof(writecache));

	bb = 0;
	ilen = 0, olen = 0, last_m_off = 1;

	for (;;)
	{
		u32 m_off, m_len;

		while (getbit(bb))
		{
//			dst[olen++] = src[ilen++];
 			WriteOffset(src[ilen++],olen++,(u8*)writecache,&writeoffset,dst,sizeof(writecache));
		}
		m_off = 1;
		do
		{
			m_off = m_off*2 + getbit(bb);
		} while (!getbit(bb));
		if (m_off == 2)
		{
			m_off = last_m_off;
		}
		else
		{
			m_off = (m_off-3)*256 + src[ilen++];
			if (m_off == u32_C(0xffffffff))
			{
				break;
			}
			last_m_off = ++m_off;
		}
		m_len = getbit(bb);
		m_len = m_len*2 + getbit(bb);
		if (m_len == 0)
		{
			m_len++;
			do
			{
				m_len = m_len*2 + getbit(bb);
			} while (!getbit(bb));
			m_len += 2;
		}
		m_len += (m_off > 0xd00);

		{
			const u8 *m_pos;
			u8 value;
			int m_off_offset;
//			m_pos = dst + olen - m_off;
//			dst[olen++] = *m_pos++;

			m_off_offset = olen - m_off;
			value = ReadOffset(m_off_offset++,(u8*)writecache,&writeoffset,dst,sizeof(writecache));
 			WriteOffset(value,olen++,(u8*)writecache,&writeoffset,dst,sizeof(writecache));
			do
			{
//				dst[olen++] = *m_pos++;
				value = ReadOffset(m_off_offset++,(u8*)writecache,&writeoffset,dst,sizeof(writecache));
	 			WriteOffset(value,olen++,(u8*)writecache,&writeoffset,dst,sizeof(writecache));
			} while (--m_len > 0);
		}
	}

	*dst_len = olen;
	FlushCache((u8*)writecache,&writeoffset,(u8*)dst,sizeof(writecache));

	return ilen == src_len ? (0/*E_OK*/) : (-1);
}

