#include "ModFile.h"

ModFile::ModFile()
{
	mGuessedInstruments = 15;
	mGuessedChannels = 4;
	mGuessedPatternOffset = 20+(15*30)+1+1+128;
	mSampleOffset = 0;
	mPatternLength = 64;
	mIsExtendedMod = false;
}

ModFile::~ModFile()
{
}

#define MAKEFORMAT(a,b,c,d) int(int(a) | (int(b)<<8) | (int(c)<<16) | (int(d)<<24))

bool ModFile::ParseMod(void)
{

	switch(GetIntNS(1080))
	{
		case MAKEFORMAT('M','.','K','.'):
		case MAKEFORMAT('F','L','T','4'):
		case MAKEFORMAT('M','!','K','!'):
		case MAKEFORMAT('4','C','H','N'):
		{
			mGuessedChannels = 4;
			mGuessedInstruments = 31;
			mGuessedPatternOffset = 1084;
			break;
		}
		case MAKEFORMAT('8','C','H','N'):
		case MAKEFORMAT('O','C','T','A'):
		case MAKEFORMAT('F','L','T','8'):
		{
			mGuessedChannels = 8;
			mGuessedInstruments = 31;
			mGuessedPatternOffset = 1084;
		}
	}

	mChannels = mGuessedChannels;

	ParseTitle();
	ParsePatternTable();
	ParsePatterns();
	ParseInstruments();

	return true;
}

bool ModFile::ParseTitle(void)
{
	mTitle = GetString(0,20);

	return 1;
}

bool ModFile::ParseInstruments(void)
{
	int total = 0;
	for (int i=0;i<mGuessedInstruments;i++)
	{
		Instrument *instr = new Instrument;
		instr->mName = GetString(20+(i*30),22);
		instr->mLength = 2*(unsigned int)GetShort(20+(i*30)+22);
		instr->mFineTune = mAddress[20+(i*30)+22+2];
		instr->mLinearVolume = mAddress[20+(i*30)+22+2+1];
		instr->mLoopStart = 2*(unsigned int)GetShort(20+(i*30)+22+2+1+1);
		instr->mLoopSize = 2*(unsigned int)GetShort(20+(i*30)+22+2+1+1+2);
		if (instr->mLength > 0)
		{
			instr->mSampleData = (char *) malloc(instr->mLength);
			memcpy(instr->mSampleData,mAddress+mSampleOffset+total,instr->mLength);
		}
		else
		{
			instr->mSampleData = 0;
		}
		mInstruments.push_back(instr);
		total += instr->mLength;
	}

	return 1;
}

bool ModFile::ParsePatternTable(void)
{
	int i;
	int offset = 20 + (mGuessedInstruments * 30);
	int numpats = mAddress[offset];

	mEndSongJumpPosition = ((unsigned char *)mAddress)[offset+1];
	if (mEndSongJumpPosition >= numpats)
	{
		mEndSongJumpPosition = 127;
	}
	for (i=0;i<numpats;i++)
	{
		mPatternsInSeq.push_back(((unsigned char *)mAddress)[offset + 2 + i]);
	}

	int highest = 0;
	for (i=0;i<128;i++)
	{
		if (((unsigned char *)mAddress)[offset + 2 + i] > highest)
		{
			highest = ((unsigned char *)mAddress)[offset + 2 + i];
		}
	}
	highest++;

	mNumberPatterns = highest;

	mSampleOffset = mGuessedPatternOffset + (highest * mGuessedChannels * 4 * 64);

	return 1;
}

bool ModFile::ParsePatterns(void)
{
	for (int i=0;i<mNumberPatterns;i++)
	{
		int *data;
		data = (int *) malloc(mGuessedChannels * 4 * 64);
		memcpy(data,mAddress + mGuessedPatternOffset + (i * mGuessedChannels * 4 * 64),(mGuessedChannels * 4 * 64));
		mPatternData.push_back(data);
	}

	return 1;
}

// Utility functions

int ModFile::GetIntNS(int offset)
{
	return MAKEFORMAT((unsigned char)mAddress[offset],(unsigned char)mAddress[offset+1],(unsigned char)mAddress[offset+2],(unsigned char)mAddress[offset+3]);
}

unsigned short ModFile::GetShort(int offset)
{
	return int(((unsigned char *)mAddress)[offset+1]) | (int(((unsigned char *)mAddress)[offset])<<8);
}

unsigned short ModFile::GetShortNS(int offset)
{
	return int(((unsigned char *)mAddress)[offset+0]) | (int(((unsigned char *)mAddress)[offset+1])<<8);
}

std::string ModFile::GetString(int offset,int maxlen)
{
	char buffer[256];
	if (maxlen >= sizeof(buffer))
	{
		return "";
	}

	memcpy(buffer,mAddress+offset,maxlen);
	buffer[maxlen] = 0;
	return std::string(buffer);
}

/*****
 * XM Parsing stuff
 */

bool ModFile::XMParseMod(void)
{
	if (strncmp(mAddress,"Extended Module",15) != 0)
	{
		return false;
	}

	mIsExtendedMod = true;

	mXMHeaderSize = GetIntNS(60)+60;

	mGuessedChannels = GetShortNS(60+8);
	mChannels = mGuessedChannels;
	mXMDefaultTempo = GetShortNS(60+16);
	mXMDefaultBPM = GetShortNS(60+18);

	mXMUseAmigaFreqTable = false;
	if ( mAddress[60+14] & 1 )
	{
		mXMUseAmigaFreqTable = true;
	}

	XMParseTitle();
	XMParsePatternTable();
	XMParsePatterns();
	XMParseInstruments();

	mExtendedData.push_back(mXMDefaultTempo);
	mExtendedData.push_back(mXMDefaultBPM);
	mExtendedData.push_back(mXMUseAmigaFreqTable);

	return true;
}

bool ModFile::XMParseTitle(void)
{
	mTitle = GetString(17,20);

	return true;
}

bool ModFile::XMParseInstruments(void)
{
	mGuessedInstruments = GetShortNS(60+12);

	int numinst = mGuessedInstruments;
	int offset = mSampleOffset;
	int realinstr=1;
	while(numinst > 0)
	{
		int instrsize = GetIntNS(offset);
		int type = mAddress[offset+26];
		int numsamp = GetShortNS(offset+27);
		// Gather suitable information for instruments
		std::string mName = GetString(offset+4,22);
		int mLength = 0;
		int mFineTune = 0;
		int mLinearVolume = 0;
		int mLoopStart = 0;
		int mLoopSize = 0;
		char *mSampleData = 0;
		int relativenote = 0;

		printf("Instrument %d ($%x) name '%s'\n",realinstr,realinstr++,&mName[0]);
		printf("Instrument size %d ($%x)\n",instrsize,instrsize);
		printf("Num samples %d\n",numsamp);

/*
  - 1st part of 1st instrumentheader
  - if nr_of_samples>0 then this data follows:
      - 2nd part of 1st instrumentheader
      - 1st sampleheader
      - 2nd sampleheader follows (if any)
       .....
      - last sampleheader (if any)
      - if the sample_size>0 (1st sample header) then sampledata of
        1st sample follows
      - if the sample_size>0 (2nd sample header) then sampledata of
        2nd sample follows
       .....
      - if the sample_size>0 (last sample header) then sampledata of
        last sample follows
*/

/*
		offset+=4+22+1+2;
		if (numsamp > 0)
		{
			int sampheadsize = GetIntNS(offset);
//			offset+=sampheadsize;
			offset+=4+96+48+48;
		}
		if (numsamp == 0)
		{
			printf("No samples\n");
		}
		offset+=11+2+(14*1);

		offset+=11;
*/

		offset+=instrsize;

		if (numsamp > 1)
		{
			printf("numsamp > 1 for this instrument\nCannot parse yet\n");
//			exit(-1);
		}
		if(numsamp > 0)
		{
			mLength = GetIntNS(offset);
			mLoopStart = GetIntNS(offset+4);
			mLoopSize = GetIntNS(offset+8);
			mLinearVolume = ((unsigned char *)mAddress)[offset+12];
			int type = ((unsigned char *)mAddress)[offset+14];
			mName = GetString(offset+18,22);
			relativenote = (int) ((signed char *)mAddress)[offset+16];
			mFineTune = (int) ((signed char *)mAddress)[offset+13];

			printf("Sample name '%s'\n",&mName[0]);

			offset+=18+22;
			numsamp--;

			int extralength = 0;
			while(numsamp)
			{
				extralength += GetIntNS(offset);
				int type = ((unsigned char *)mAddress)[offset+14];
				offset+=18+22;
				numsamp--;
			}

			if (type & 16)
			{
				mSampleData = (char *) calloc(mLength/2,1);
				signed short old;
				old=0;
				for (int i=0 ;i < mLength/2;i++)
				{
					signed short news;
					news=((signed short*)(mAddress+offset))[i]+old;
					old=news;
					mSampleData[i] = (char) (news/256);
				}
				offset+=mLength;
				mLength /= 2;
				mLoopStart /= 2;
				mLoopSize /= 2;
			}
			else
			{
				mSampleData = (char *) calloc(mLength,1);
				signed char old;
				old=0;
				for (int i=0 ;i < mLength;i++)
				{
					signed char news;
					news=((signed char *)(mAddress+offset))[i]+old;
					old=news;
					mSampleData[i] = (char) news;
				}
				offset+=mLength;
			}

			printf("Sample type = %d\n",type);
			printf("Sample length = %d ($%x)\n",mLength,mLength);
			printf("relative note = %d\n",relativenote);
			printf("finetune = %d\n",mFineTune);

			offset+= extralength;
		}

		Instrument *instr = new Instrument;
		instr->mName = mName;
		instr->mLength = mLength;
		instr->mFineTune = mFineTune;
		instr->mLinearVolume = mLinearVolume;
		instr->mLoopStart = mLoopStart;
		instr->mLoopSize = mLoopSize;
		instr->mRelativeNote = relativenote;
		if (instr->mLength > 0)
		{
			instr->mSampleData = mSampleData;
		}
		else
		{
			instr->mSampleData = 0;
		}
		mInstruments.push_back(instr);

		numinst--;
	}

	return true;
}

bool ModFile::XMParsePatternTable(void)
{
	int i;
	int offset = 60+20;
	int numpats = GetShortNS(60+4);

	mEndSongJumpPosition = GetShortNS(60+6);
	if (mEndSongJumpPosition >= numpats)
	{
		mEndSongJumpPosition = 511;
	}
	for (i=0;i<numpats;i++)
	{
		mPatternsInSeq.push_back(((unsigned char *)mAddress)[offset+i]);
	}

	mNumberPatterns = GetShortNS(60+10);

	return true;
}

#pragma pack(1)
typedef struct s_XMNoteFormat
{
	unsigned char mNote;
	unsigned char mInstrument;
	unsigned char mVolume;
	unsigned char mEffectType;
	unsigned char mEffectParameter;
} XMNoteFormat;
#pragma pack()

int PeriodTab[] = {
	907,900,894,887,881,875,868,862,856,850,844,838,832,826,820,814,
	808,802,796,791,785,779,774,768,762,757,752,746,741,736,730,725,
	720,715,709,704,699,694,689,684,678,675,670,665,660,655,651,646,
	640,636,632,628,623,619,614,610,604,601,597,592,588,584,580,575,
	570,567,563,559,555,551,547,543,538,535,532,528,524,520,516,513,
	508,505,502,498,494,491,487,484,480,477,474,470,467,463,460,457
	};

bool ModFile::XMParsePatterns(void)
{
	int offset = mXMHeaderSize;
	int numpats = mNumberPatterns;

	while(numpats > 0)
	{
		int headsize = GetIntNS(offset);
		int packtype = mAddress[offset+4];
		if (packtype != 0)
		{
			printf("Packtype != 0\nCannot parse\n");
			exit(-1);
		}
		int patrows = GetShortNS(offset+5);
		int packsize = GetShortNS(offset+7);
		XMNoteFormat *patterndata = (XMNoteFormat*) calloc(mChannels*patrows,sizeof(XMNoteFormat));

		offset+=headsize;

		int noteoffset = 0;
		int j = 0;
		unsigned char *src = (unsigned char *) mAddress + offset;

		XMNoteFormat *note;

		for (int row = 0;row<patrows;row++)
		{
			for (int channel = 0;channel<mChannels;channel++)
			{
				if (j < packsize)
				{
					note = &patterndata[channel + (row*mChannels)];
					unsigned char b = src[j++];
					if (b & 0x80)
					{
						if (b & 1) note->mNote = src[j++];
						if (b & 2) note->mInstrument = src[j++];
						if (b & 4) note->mVolume = src[j++];
						if (b & 8) note->mEffectType = src[j++];
						if (b & 16) note->mEffectParameter = src[j++];
					}
					else
					{
						note->mNote = b;
						note->mInstrument = src[j++];
						note->mVolume = src[j++];
						note->mEffectType = src[j++];
						note->mEffectParameter = src[j++];
					}
				}
			}
		}

		/* Convert the XM data to a mod format */
		int *oldstylepatdata = (int *) calloc(mChannels*patrows,4);

		for (row = 0;row<patrows;row++)
		{
			for (int channel = 0;channel<mChannels;channel++)
			{
				note = &patterndata[channel + (row*mChannels)];

/*
           Byte  1   Byte  2   Byte  3   Byte 4
          --------- --------- --------- ---------
          7654-3210 7654-3210 7654-3210 7654-3210
          wwww XXXX xxxxxxxxx yyyy ZZZZ zzzzzzzzz

              wwwwyyyy ( 8 bits) : sample number
          XXXXxxxxxxxx (12 bits) : sample 'period'
          ZZZZzzzzzzzz (12 bits) : effect and argument
*/

				if (note->mNote >= 1 && note->mNote <= 97)
				{
					int period = note->mNote;

					oldstylepatdata[channel + (row*mChannels)]  = (period&255)<<8;
					oldstylepatdata[channel + (row*mChannels)] |= ((period>>8)&255);

					oldstylepatdata[channel + (row*mChannels)] |= (note->mInstrument&15)<<(8+8+4);
					oldstylepatdata[channel + (row*mChannels)] |= ((note->mInstrument>>4)&15)<<(4);

					if (note->mEffectType == 0x0d)
					{
						oldstylepatdata[channel + (row*mChannels)] |= 0x0d;
					}
				}
			}
		}

		mPatternData.push_back(oldstylepatdata);

		printf("Pattern %d ($%x) is %d rows long\n",mPatternLengths.size(),mPatternLengths.size(),patrows);

		mPatternLengths.push_back(patrows);

		free(patterndata);

		offset += packsize;

		numpats--;
	}

// Update this
	mSampleOffset = offset;
	return true;
}

