#include <stdio.h>
#include <stdlib.h>
#include "ModFile.h"
#include <windows.h>
#include "DataBase.h"

int main(int argc,char **argv)
{
	int i;

	if (argc < 5)
	{
		printf("Invalid number of params\n");
		printf("eg: <modfile>.MOD <asmfile>.s <headerfile>.h <leafname> [-n] [-c]\n");
		printf("-n   Create a new sample database. Implies -c\n");
		printf("-c   Use the sample database when outputting sample names\n");
		exit(-1);
	}

	DataBase database;

	if (argc >= 6)
	{
		if (argv[5][1] == 'n')
		{
			database.Clear();
		}
		else
		{
			database.Load();
		}
	}

	FILE *fp;
	char *moddata;

	fp = fopen(argv[1],"rb");

	if (!fp)
	{
		printf("Modfile '%s' not found!\n",argv[1]);
		exit(-1);
	}

	fseek(fp,0,SEEK_END);
	int size = ftell(fp);
	fseek(fp,0,SEEK_SET);

	moddata = (char *) malloc(size);
	if (!moddata)
	{
		printf("Out of memory\n");
		exit(-1);
	}

	fread(moddata,1,size,fp);
	fclose(fp);

	ModFile theMod;
	theMod.SetAddress(moddata);
	char buffer[256];
	strcpy(buffer,argv[1]);
	for (i=0;i<strlen(buffer);i++)
	{
		buffer[i] = tolower(buffer[i]);
	}

	if (strstr(buffer,".mod"))
	{
		theMod.ParseMod();
	}
	else if (strstr(buffer,".xm"))
	{
		theMod.XMParseMod();
	}


	SetFileAttributes(argv[2],FILE_ATTRIBUTE_NORMAL);
	DeleteFile(argv[2]);
	SetFileAttributes(argv[3],FILE_ATTRIBUTE_NORMAL);
	DeleteFile(argv[3]);

	fp = fopen(argv[2],"w");
	if (!fp)
	{
		printf("Error opening %s for output\n",argv[2]);
		exit(-1);
	}

	if (argc >= 6)
	{
		fprintf(fp,"\x09.text\n\x09.global\x09g%sSamples\n\x09.type\x09g%sSamples, object\n\x09.align	4\ng%sSamples:\n",argv[4],argv[4],argv[4]);
		int i;
		for (i = 0;i<theMod.mInstruments.size();i++)
		{
			DataBase::ItemData *item;
			int found;
			item = new DataBase::ItemData;
			item->mName = theMod.mInstruments[i]->mName;
			item->mSize = theMod.mInstruments[i]->mLength;
			item->mData = theMod.mInstruments[i]->mSampleData;
			found = database.Find(item);
			if (found < 0)
			{
//				printf("Adding new sample of %d bytes\n",item->mSize);
				item = new DataBase::ItemData;
				item->mName = theMod.mInstruments[i]->mName;
				item->mSize = theMod.mInstruments[i]->mLength;
				item->mData = theMod.mInstruments[i]->mSampleData;
				database.Add(item);

				item = new DataBase::ItemData;
				item->mName = theMod.mInstruments[i]->mName;
				item->mSize = theMod.mInstruments[i]->mLength;
				item->mData = theMod.mInstruments[i]->mSampleData;
				found = database.Find(item);
			}
			else
			{
//				printf("Found a sample of %d bytes handle %d\n",item->mSize,found);
			}
			fprintf(fp,"\x09.word gGlobalSampleData_%d,%d,%d,%d,%d,%d\n",found,theMod.mInstruments[i]->mLength,theMod.mInstruments[i]->mFineTune,theMod.mInstruments[i]->mLinearVolume,theMod.mInstruments[i]->mLoopStart,theMod.mInstruments[i]->mLoopSize);
		}

		DataBase::ItemData *item;
		database.Iterate(true);
		int handle = 0;
		SetFileAttributes(argv[6],FILE_ATTRIBUTE_NORMAL);
		DeleteFile(argv[6]);

		FILE *fp = fopen(argv[6],"w");
		if (!fp)
		{
			printf("Cannot write the global sample file\n");
			exit(-1);
		}
		while ( (item = database.Iterate()) != 0)
		{
			fprintf(fp,"\n");
			fprintf(fp,"\x09.text\n\x09.global\x09gGlobalSampleData_%d\n\x09.type\x09gGlobalSampleData_%d, object\n\x09.align	4\n",handle,handle);
			fprintf(fp,"gGlobalSampleData_%d:\n",handle);
			int j,k=0;
			for (j=0;j<item->mSize;j+=4,k--)
			{
				if (k == 0)
				{
					k = 8;
					fprintf(fp,"\n\x09.word ");
				}
				else
				{
					if (j <= item->mSize-1)
					{
						fprintf(fp,",");
					}
				}
				fprintf(fp,"0x%8.8x",((int *)item->mData)[j/4]);
			}
			fprintf(fp,"\n\n");
			handle++;
		}
		fclose(fp);
	}
	else
	{
		fprintf(fp,"\x09.text\n\x09.global\x09g%sSamples\n\x09.type\x09g%sSamples, object\n\x09.align	4\ng%sSamples:\n",argv[4],argv[4],argv[4]);
		int i;
		for (i = 0;i<theMod.mInstruments.size();i++)
		{
			fprintf(fp,"\x09.word %sSampleData_%d,%d,%d,%d,%d,%d\n",argv[4],i,theMod.mInstruments[i]->mLength,theMod.mInstruments[i]->mFineTune,theMod.mInstruments[i]->mLinearVolume,theMod.mInstruments[i]->mLoopStart,theMod.mInstruments[i]->mLoopSize);
		}

		for (i = 0;i<theMod.mInstruments.size();i++)
		{
			fprintf(fp,"\n");
			fprintf(fp,"%sSampleData_%d:\n",argv[4],i);
			int j,k=0;
			for (j=0;j<theMod.mInstruments[i]->mLength;j+=4,k--)
			{
				if (k == 0)
				{
					k = 8;
					fprintf(fp,"\n\x09.word ");
				}
				else
				{
					if (j <= theMod.mInstruments[i]->mLength-1)
					{
						fprintf(fp,",");
					}
				}
				fprintf(fp,"0x%8.8x",((int *)theMod.mInstruments[i]->mSampleData)[j/4]);
			}
			fprintf(fp,"\n\n");
		}
	}

	printf("Number of instruments used %d\n",theMod.mInstruments.size());

	fprintf(fp,"\x09.text\n\x09.global\x09g%sPatternList\n\x09.type\x09g%sPatternList, object\n\x09.align	4\ng%sPatternList:\n",argv[4],argv[4],argv[4]);
	fprintf(fp,"\x09.word %d\n",theMod.mPatternsInSeq.size());
	fprintf(fp,"\x09.word %d\n",theMod.mEndSongJumpPosition);
	fprintf(fp,"\x09.word %d\n",theMod.mChannels);
	for (i=0;i<theMod.mPatternsInSeq.size();i++)
	{
		fprintf(fp,"\x09.word %sPatternData%d\n",argv[4],theMod.mPatternsInSeq[i]);
	}
	fprintf(fp,"\n");

	for (i=0;i<theMod.mPatternData.size();i++)
	{
		fprintf(fp,"%sPatternData%d:\n",argv[4],i);
		int j,k=0;
		int numrows = 64;
		if (theMod.mIsExtendedMod)
		{
			numrows = theMod.mPatternLengths[i];
		}
		for (j=0;j<(theMod.mChannels * 4 * numrows);j+=4,k--)
		{
			if (k == 0)
			{
				k = theMod.mChannels;
				fprintf(fp,"\n\x09.word ");
			}
			else
			{
				if (j < (theMod.mChannels * 4 * numrows))
				{
					fprintf(fp,",");
				}
			}
			fprintf(fp,"0x%8.8x",((int *)theMod.mPatternData[i])[j/4]);
		}
		fprintf(fp,"\n\n");
	}

	if (theMod.mIsExtendedMod)
	{
		fprintf(fp,"\x09.text\n\x09.global\x09s%sExtendedPatternData\n\x09.type\x09s%sExtendedPatternData, object\n\x09.align	4\ns%sExtendedPatternData:\n",argv[4],argv[4],argv[4]);
		fprintf(fp,"\x09.word %d\n",theMod.mPatternLengths.size());
		for (i = 0;i<theMod.mPatternLengths.size();i++)
		{
			fprintf(fp,"\x09.word %d\n",theMod.mPatternLengths[i]);
		}
		fprintf(fp,"\n");

		fprintf(fp,"\x09.text\n\x09.global\x09s%sExtendedSampleData\n\x09.type\x09s%sExtendedSampleData, object\n\x09.align	4\ns%sExtendedSampleData:\n",argv[4],argv[4],argv[4]);
		for (i = 0;i<theMod.mInstruments.size();i++)
		{
			fprintf(fp,"\x09.word %d,0,0,0,0,0,0,0\n",theMod.mInstruments[i]->mRelativeNote);
		}
		fprintf(fp,"\n");

		fprintf(fp,"\x09.text\n\x09.global\x09g%sExtended\n\x09.type\x09g%sExtended, object\n\x09.align	4\ng%sExtended:\n",argv[4],argv[4],argv[4]);
		fprintf(fp,"\x09.word s%sExtendedPatternData\n",argv[4]);
		fprintf(fp,"\x09.word s%sExtendedSampleData\n",argv[4]);
		for (i=0;i<theMod.mExtendedData.size();i++)
		{
			fprintf(fp,"\x09.word %d\n",theMod.mExtendedData[i]);
		}
		fprintf(fp,"\n");
	}

	fclose(fp);

	fp = fopen(argv[3],"w");
	if (!fp)
	{
		printf("Error opening %s for output\n",argv[2]);
		exit(-1);
	}

	fprintf(fp,"extern u32 g%sSamples[];\n",argv[4]);
	fprintf(fp,"extern u32 g%sPatternList[];\n",argv[4]);
	if (theMod.mIsExtendedMod)
	{
		fprintf(fp,"extern u32 g%sExtended[];\n",argv[4]);
	}
	fclose(fp);

	if (argc >= 6)
	{
		database.Save();
	}

	return 0;
}
