#ifdef WIN32
#include <stdio.h>
#include <stdlib.h>
#endif
#include "RawDecompress.h"
#include "GBADecompress.h"

// The decompression routine

/**
 * This decompress map data saved from the editor for use with the GBA
 * \param input the input data for the map
 * \param output the place to decompress the map to, pass in NULL to have it allocated
 * \param width the width to fill in, pass NULL to ignore this
 * \param height the height to fill in, pass NULL to ignore this
 * \return the pointer for the map
 */
s16 *TestDecompress(u8 *input,s16 *output,s16 *width,s16 *height)
{
	s16 *data;
	s32 wid,hei,i;
	u8 *layerdata;
	u8 *stripdata;
	s32 srclen;
	u32 destlen;

	data = output;

	wid = input[0] | (input[1]<<8);
	hei = input[2] | (input[3]<<8);
	layerdata = input + 2 + 2 +4;

#ifdef WIN32
	// For the PC we malloc some memory if needed
	if (data == NULL)
	{
		data = (s16 *) malloc(wid*hei*2);
	}
#endif

	srclen = input[4] | (input[5]<<8) | (input[6]<<16) | (input[7]<<24);

	GBADecompress(layerdata,srclen,(u8 *) data,&destlen,0);

	// Fill in the width and height in the pointers passed in are not NULL
	if (width != 0)
	{
		*width = wid;
	}

	if (height != 0)
	{
		*height = hei;
	}

	// Return the pointer to the data
	return data;
}

