/* 
	GBA Particle system

	uses 8x8 pixel sprites (single character) per particle.
	reserve 32 sprites from the OAM for particles.
	switch on alpha for the sprite layer (so particles blend to each other + sprites)

	particles will have the following parameters

	sprite number
	sprite character (stored in sprite)
	xpos (stored in sprite)
	ypos (stored in sprite)
	dx
	dy
	gravity x
	gravity y
	decay time

*/

//theoretically, we can spawn particles in various ways, including creating them as larger graphics than 8x8
//as the particle handler is only worried about moving them about.

#include <AGB.h>
#include "asldefines.h"
//#include "data.h"
#include "particles.h"
#include "particlesprites.h"
#include "spritelist.h"
#include "flash.h"
#include "main.h"
#include "raineffect.h"
#include "powerupfx.h"
#include "player.h"
#include "baddydeath.h"
#include "torchcover.h"
#include "darkcover.h"
#include "Frisbee.h"

extern s16 game_x;
extern s16 game_y;
extern u8 currsprite;
extern u8 elem_power_level;

u8 torchactive;
u8 TorchLitO;

s16 rainswitchtlx[8];
s16 rainswitchtly[8];
s16 rainswitchbrx[8];
s16 rainswitchbry[8];
s8 rainactive;

particle_type particle[64];
trailtype trail[16];
blasttype Blast[16];

u8 flash_frame;
u8 flash_base;
u8 flashframeclock;
u8 numparticles;
u8 particleclock;

s16 flash_x,flash_y;
u8 flash_scale;
u8 flash_scroll_x,flash_scroll_y;
u8 power_scale;
s8 power_sign;

s8 rain_x,rain_y;

u8 rainontime;
u8 nxttrail;
u8 nxtparticle;
u8 LastParticles; 
u16 raindepth;

u8 af;
u8 sc;
u8 particle_toggle;

const u16 RainColour[4] = {0x62d4,0x6290,0x520e,0x49cc}; // 128 160 192		TextColour(15,20,25);

const u32 PowerupSprite[3] = {(u32)PowerupFXSprite,(u32)PowerupOrbSprite,(u32)PowerupSplashSprite};

const u8 pulse[16] = {4,3,2,1,1,2,3,2,1,2,3,4,5,6,7,8};

void initrainer(void);
void initbaddyblast(void);
void HandleRainer(void);
void inittorch(void);
void HandleTorch(void);

void HandleFlash(void);
void HandleBlast(void);
void HandlePlayerPower(void);


void InitParticles(void)
{
	numparticles = 0;
	particleclock = 0;
	LastParticles = 0;
	nxtparticle = 0;
}

u16 spawnparticle(s16 xpos,s16 ypos,s8 dx,s8 dy,s8 grav_x,s8 grav_y,u8 time,u8 charno,u8 fx)
{
	u8 frame;
	s8 a;
	
	
	if(xpos < game_x || xpos > game_x + 240)
		return;
	
	if(numparticles >= 64)
		numparticles = 63;
	
	a = numparticles++;	

//	if(numparticles == 32)
//		numparticles = 31; // always generate a new particle, even if it overwites the last one




	frame = 96 + a / 2;
 	//Sprite[frame] = *(OamData *)OamData_Sample;
  	particle[a].sprite = *(OamData *)OamData_Sample;
   	particle[a].sprite.Pltt = 15;
	particle[a].sprite.CharNo = PARTICLE_SPRITE;
   	particle[a].sprite.AffineMode = 0;	
	
	if(charno >= 48 && charno < 50)
		charno -= 16;
	if(charno < 32)
	{
		charno = charno << 2; // normal particle
   		particle[a].sprite.Shape = 0;
   		particle[a].sprite.Size = 0; // 8 x 8
		particle[a].sprite.ObjMode = 0;
	}
	else
 	{ 	 
		// flame, smoke etc.
 		if(charno == 50) // smoke
 		{
 			charno = 64;
 			particle[a].sprite.Shape = 0;
 			particle[a].sprite.Size = 2; // 32 x 32
 			particle[a].sprite.ObjMode = 0;
 			particle[a].sprite.CharNo = SMOKE_SPRITE;
   			particle[a].sprite.HFlip = 0;
   			particle[a].sprite.AffineMode = 1;
   			particle[a].sprite.AffineParamNo_L = 15 + 4; // use affine 4 + scale
			SpriteScaleFine(8, 18 - 8, 18 - 8);
		}
 		else
 		{
 			particle[a].sprite.Shape = 0;
 			particle[a].sprite.Size = 1; // 16 x 16
 			particle[a].sprite.ObjMode = 0;
 		}
 	}
   	
   	if(fx & 2)
   		particle[a].sprite.ObjMode = 2; // window
   		
    particle[a].sprite.Priority = 1;
   	if(fx & 4) 
   		particle[a].sprite.Priority = 3; // in bg
	//particle[a].sprite = Sprite[frame];
	particle[a].charno = charno;
	particle[a].xpos = xpos << 3;
	particle[a].ypos = ypos << 3;
	particle[a].dx = dx;
	particle[a].dy = dy;
	particle[a].grav_x = grav_x;
	particle[a].grav_y = grav_y;
	particle[a].time = time;
	particle[a].fx = fx;
	particle[a].t = particle_toggle;
	particle_toggle = 1 - particle_toggle;
	return frame;
}

void DoParticles(void)
{
	u8 a;
	u8 frame;
	s16 sprx,spry;
	u32 dest,length;
	
	handletrailsprites();
	
	particleclock++;
	if(particleclock == 9)
		particleclock = 0;
	
	dest = &Sprite[96];
	length = 32 * sizeof(OamData);
	DmaClear(3,160,dest,length,32);
		
	for(a = 0; a < numparticles; a++)
	{
		frame = 96 + (a >> 1);
		if(particle[a].time > 0)
		{

			sprx = (particle[a].xpos >> 3) - game_x;
			spry = (particle[a].ypos >> 3) - game_y;
			if(sprx >= -64 && sprx < 280 && spry >= -64 && spry <= 192)
			{
				
				if((a & 1) == (GameClock & 1))
				{
					Sprite[frame] = particle[a].sprite;
					Sprite[frame].HPos = sprx & 511;
					Sprite[frame].VPos = (spry + 256) & 255;
					Sprite[frame].CharNo = PARTICLE_SPRITE + particle[a].charno;
					if(particle[a].charno < 64)
					{
#if 1					
						if(particle[a].charno > 31 ) // flames
						{
							if(particleclock == 0)
							{
								if(rndfast(1))
									particle[a].charno = 32;
								else
									particle[a].charno = 36;
							}
						}
#endif
					}
					else
					{
					//	frame = 96 + a;
			   			sc = min(15,particle[a].time);
						af = 3 - (sc / 4);
						Sprite[frame].CharNo = SMOKE_SPRITE + af * 16;
			   			//Sprite[frame].HFlip = 0;
			   			af = 15 + sc;
			   			Sprite[frame].AffineMode = 0;
			   			
			   			Sprite[frame].VFlip = 0; //(af >> 4) & 1;
			   			Sprite[frame].HFlip = 0;//(af >> 3)  & 1;
			   			Sprite[frame].AffineParamNo_L = 0; //af & 7 ; // use affine 4 + scale
				//		SpriteScaleFine(af,72 - sc * 4, 72 - sc * 4);
				//		sprx -= 16;
				//		spry -= 16;
						
						Sprite[frame].HPos = sprx & 511;
						Sprite[frame].VPos = (spry + 256) & 255;
					//	SpriteScaleFine(4 + particle[a].time ,  particle[a].time,  particle[a].time);
					}
				}
				if(particle[a].fx & 1) // leaf spiral fall
				{
					if(particleclock == 0)
					{	particle[a].charno++;
						if((particle[a].charno & 3) == 0)
							particle[a].charno = particle[a].charno - 4;
					 }
					 if(particle[a].dy > 0)
					 {
						particle[a].xpos += fsin[(u8)(particle[a].grav_x + 128)] >> 4;
						particle[a].grav_x+=4;
						if(particle[a].dy > 8)
							particle[a].dy = 8;
					}
					else
					{
						particle[a].xpos += particle[a].dx;
					}
				}
				else
				{ 
					particle[a].xpos += particle[a].dx;
					if(particle[a].grav_x == -99)
						particle[a].dx += srnd(4);
					else
						particle[a].dx += particle[a].grav_x;
					if(particle[a].time < 8 && particle[a].charno < 28)
						particle[a].charno = (particle[a].charno & 60) + 3 - ( particle[a].time >> 1); // fade off
				}
				// 
				//
				if((a & 1) == (GameClock & 1))
					Sprite[frame].VPos = (spry + 256) & 255;
				particle[a].ypos += particle[a].dy;

				particle[a].dy += particle[a].grav_y;

				particle[a].time--;
			}
			else
			{
				particle[a].time = 0;
			}
		}
		else
		{
			frame = numparticles - 1;
			if(a < frame)
			{
				particle[a] = particle[frame];
			}

			a--;
			numparticles--;
		}

		
	}
//	for(a = numparticles; a < LastParticles;a+= 2)
//	{
	//	if((a & 1) == (GameClock & 1))
//			Sprite[96 + (a >> 1)].VPos = 160;
//	}
#if 0	
	if(LastParticles > numparticles) // If less sprites assigned this frame, switch off the remainder
	{
		u32 dest;
		u32 length;
		dest = &Sprite[numparticles + 96];
		length= (LastParticles - numparticles) * 8;
		DmaClear(3,161,dest,length,32)
	}
#endif
	LastParticles = numparticles;
	
}

// some nice standard particle spawners

void SpawnParticleSplash(s16 xpos,s16 ypos,u8 charno)
{
	u8 a;
	for(a = 0; a < 8; a++)
		spawnparticle(xpos,ypos,srnd(8),-rnd(32),0,1,32,charno,0);

}

void SpawnParticleLeafSplash(s16 xpos,s16 ypos,u8 charno)
{
	u8 a;
	for(a = 0; a < 8; a++)
		spawnparticle(xpos,ypos,srnd(8),-rnd(8)-8,rnd(127),1,64,charno,1);


}

void SpawnParticleBlast(s16 xpos,s16 ypos)
{
	u8 a;
	for(a = 0 ; a < 4 ; a++)
		spawnparticle(xpos,ypos,fsin[(a * 64 + 96) & 255] >> 3,fsin[a * 64 + 32] >> 3,0,0,64,1,0);
}

void SpawnParticleDriftUp(s16 xpos,s16 ypos,u8 charno)
{
	u8 a;
	for(a = 0; a < 16;a ++)
		spawnparticle(xpos ,ypos + srnd(8),0,-rnd(12)-2,-99,0,64,charno,0);
}

void SpawnParticleRing(s16 xpos,s16 ypos,s8 dx,s8 dy,u8 charno)
{
	u8 a;
	
	for(a = 0; a < 16;a++)
		spawnparticle(xpos,ypos,(fsin[(a * 16 + 64) & 255] * dx) >> 6,(fsin[a * 16] * dy) >> 6,0,0,32,charno,0);
	
}

void SpawnParticleInverseRing(s16 xpos,s16 ypos,s8 dx,s8 dy,u8 charno)
{
	u8 a;
	s16 sx,sy;
	for(a = 0; a < 16;a++)
	{
		sx = ((fsin[(a * 16 + 64) & 255] * dx) >> 6);
		sy = ((fsin[a * 16] * dy) >> 6);
		spawnparticle(xpos + sx * 4,ypos + sy * 4,-sx,-sy,0,0,16,charno,0);
	}
}


void initflash(void)
{
	u8 myframe;
	myframe = GrabASprite();
	Sprite[myframe].ObjMode = 0; // window
    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
   	Sprite[myframe].CharNo = FLASH_SPRITE;
   	Sprite[myframe].Shape = 0;
   	Sprite[myframe].Size = 3;
    Sprite[myframe].Priority = 0;
   	Sprite[myframe].Pltt = 7;
   	Object[num_strats].parameter = myframe;
   	Object[num_strats].handle = HandleFlash;
   	Object[num_strats].remover = 0;
   	flash_x = parm_xpos;
   	flash_y = parm_ypos;
   	flash_scale = 1;
		SpriteScale(0,flash_scale * 2  ,max(1,flash_scale >> 2));

	flash_frame = 0;
	flashframeclock = 0;   	
	flash_base = 0; // 0 = growing, 1 = hold buildup, 2 = explode
}



void HandleFlash(void)
{
	u8 myframe,b;
	s16 sprx,spry;
	
	myframe = Object[current_strat].parameter;
	
	if(Sprite[player.frame].HFlip == 0)
		sprx = flash_x - game_x - 64 - 8;
	else
		sprx = flash_x - game_x - 64 + 8;
	if(flash_scale > 8)
		sprx += srnd(16);	
	if(flash_scale < 4)
		spry = min(player.ypos,flash_y  - 64 + flash_scale * 2) - game_y;
	else
		spry = player.ypos - 64 - game_y;
	flash_scroll_x = (flash_scroll_x + 8) & 255;
	flash_scroll_y = (flash_scroll_y + 12) & 255;
	flashframeclock++;
	b = 4;
//	if(flash_base == 2)
//		b = 8;
	if(flashframeclock >= b)
	{
		flashframeclock = 0;
		flash_frame++;
		if(flash_frame == 12)
		{
			flash_frame = 10;
		}
		if(flash_base == 1 && flash_frame > 7)
		{
			flash_frame = 0;
		}
	}
	if(sprx > -64 && sprx < 240 && spry > -64 && spry < 160 && (GameClock & 1) )
	{
   		Sprite[myframe].CharNo = FLASH_SPRITE;
		Sprite[myframe].HPos = sprx & 511;
		Sprite[myframe].VPos = (spry + 256) & 255;
		Sprite[myframe].HFlip = 0;
		Sprite[myframe].VFlip = 0;
	    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
	    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
		
		
		LoadSprite((u32)(PowerupSprite[flash_base] + (flash_frame * FLASH_SPRITE_SIZE)),FLASH_SPRITE_OFFSET,FLASH_SPRITE_SIZE);
		LoadEnemyPalette(PowerupFXPalette[min(3,elem_power_level / 60)],nxtenemypal);
		Sprite[myframe].Pltt = 2 + PalCache;
		if(flash_scale < 4 && elem_power_level < 60 )
		{
			SpriteScale(0,8  ,4);
		//	BlendUp(12 + (flash_scale / 5));
		}
		else
		{
			b = 4 + (elem_power_level / 60);
			SpriteScale(0,b,b);
		//	SpriteRotateScale(0, ((GameClock << 3) & 255),4 + (elem_power_level >> 7),4 + (elem_power_level >> 7));
		//	BlendUp(currsprite * 2 + 5);
		}
	//	spawnparticle(flash_x,flash_y,srnd(16),srnd(16),0,0,8,8 + rnd(2) * 4,4);	


	}
	else
	{
		Sprite[myframe].VPos = 160;
		SpriteScaleFine(0,1,1);
	}
	if(flash_scale > 4 || flash_base == 2)
	{
		if(flash_base < 2)
		{
			flash_frame = 0;
			flash_base = 2;
			flashframeclock = 0;
			CameraShake = 30;
			
		}
			b = 4 + (elem_power_level / 60);
			SpriteScale(0,b,b);
		if(flash_frame > 5)
		{
			//BlendAlpha(5,8);
			Sprite[myframe].VPos = 161;
			Sprite[myframe].AffineMode = 0;
			remove();
			flash_x = 0;
		}
	}
	if(player_state != player_state_special && flash_base < 2)
	{
		Sprite[myframe].VPos = 161;
		Sprite[myframe].AffineMode = 0;
		remove();
		flash_x = 0;
	}
}

void initdiskpower(void)
{
	u8 myframe;
	myframe = GrabASprite();
	Sprite[myframe].ObjMode = 0; // window
    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
   	Sprite[myframe].CharNo = FRISBEE_SPRITE;
   	Sprite[myframe].Shape = FrisbeeSizeShape[1];
   	Sprite[myframe].Size = FrisbeeSizeShape[0];
    Sprite[myframe].Priority = 0;
   	Sprite[myframe].Pltt = 7;
   	Object[num_strats].parameter = myframe;
   	Object[num_strats].handle = HandleDiskPower;
   	Object[num_strats].remover = 0;
	power_scale = 16;
	SpriteScale(0,4 + pulse[power_scale] , 4 + pulse[power_scale]);

}



void HandleDiskPower(void)
{
	u8 myframe;
	s16 sprx,spry;
	
	myframe = Object[current_strat].parameter;
	sprx = Frisbee[0].xpos - 30 - game_x;
	spry = Frisbee[0].ypos - 15 - game_y;
	if(sprx > -64 && sprx < 240 && spry > -64 && spry < 160 && (GameClock & 1))
	{
   		Sprite[myframe].CharNo = FRISBEE_SPRITE;
		Sprite[myframe].HPos = sprx & 511;
		Sprite[myframe].VPos = (spry + 256) & 255;
		Sprite[myframe].HFlip = 0;
		Sprite[myframe].VFlip = 0;
	    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
	    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
		
		
		LoadEnemyPalette( VFlashPalette,nxtenemypal);
		Sprite[myframe].Pltt = 2 + PalCache;
		SpriteScale(0,4 + pulse[power_scale] , 4 + pulse[power_scale]);
	
	}
	else
	{
		Sprite[myframe].VPos = 160;
		SpriteScaleFine(0,1,1);
	}
//	if(power_scale > )
	{
		//if((GameClock & 1) == 0)
		power_scale--;
		if(power_scale <= 0)
		{
			Sprite[myframe].VPos = 161;
			Sprite[myframe].AffineMode = 0;
			remove();
		}
	}
}


void initplayerpower(void)
{
	u8 myframe;
	myframe = GrabASprite();
	power_sign = 1;
	if(Object[num_strats].parameter)
		power_sign = -1;
	Sprite[myframe].ObjMode = 0; // window
    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
   	Sprite[myframe].CharNo = PLAYER_SPRITE;
   	Sprite[myframe].Shape = 0;
   	Sprite[myframe].Size = 3;
    Sprite[myframe].Priority = 2;
   	Sprite[myframe].Pltt = 7;
   	Object[num_strats].parameter = myframe;
   	Object[num_strats].handle = HandlePlayerPower;
   	Object[num_strats].remover = 0;
	power_scale = 32;
	SpriteScale(0,power_sign * (power_scale >> 1) ,power_scale >> 1);

}



void HandlePlayerPower(void)
{
	u8 myframe;
	s16 sprx,spry;
	
	myframe = Object[current_strat].parameter;
	sprx = player.xpos - 64 - game_x;
	spry = player.ypos - 80 - game_y + player.yposoffset;
	if(sprx > -64 && sprx < 240 && spry > -64 && spry < 160)
	{
   		Sprite[myframe].CharNo = PLAYER_SPRITE;
		Sprite[myframe].HPos = sprx & 511;
		Sprite[myframe].VPos = (spry + 256) & 255;
		Sprite[myframe].HFlip = 0;
		Sprite[myframe].VFlip = 0;
	    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
	    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
		
		
		LoadEnemyPalette( VFlashPalette,nxtenemypal);
		Sprite[myframe].Pltt = 2 + PalCache;
		SpriteScale(0,power_sign * (power_scale >> 1),power_scale >> 1);
	
	}
	else
	{
		Sprite[myframe].VPos = 160;
		SpriteScaleFine(0,1,1);
	}
	if(power_scale > 4)
	{
		//if((GameClock & 1) == 0)
		power_scale--;
		
		if(power_scale <= 4)
		{
			Sprite[myframe].VPos = 161;
			Sprite[myframe].AffineMode = 0;
			remove();
		}
	}
}

void initbosspower(void)
{
	u8 myframe;
	myframe = GrabASprite();
	Sprite[myframe].ObjMode = 0; // window
    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
   	Sprite[myframe].CharNo = BOSS_SPRITE1 + 64;
   	Sprite[myframe].Shape = 0;
   	Sprite[myframe].Size = 3;
    Sprite[myframe].Priority = 2;
   	Sprite[myframe].Pltt = 7;
   	Object[num_strats].parameter = myframe;
   	Object[num_strats].handle = HandleBossPower;
   	Object[num_strats].remover = 0;
	power_scale = 32;
	SpriteScale(0,power_scale >> 1 ,power_scale >> 1);

}



void HandleBossPower(void)
{
	u8 myframe;
	s16 sprx,spry;
	
	myframe = Object[current_strat].parameter;
	sprx = Boss_xpos - 64 - game_x;
	spry = Boss_ypos - 32 - game_y;
	if(sprx > -64 && sprx < 240 && spry > -64 && spry < 160)
	{
   		Sprite[myframe].CharNo = BOSS_SPRITE1 + 64;
		Sprite[myframe].HPos = sprx & 511;
		Sprite[myframe].VPos = (spry + 256) & 255;
		Sprite[myframe].HFlip = 0;
		Sprite[myframe].VFlip = 0;
	    Sprite[myframe].AffineMode = 3; // off 3; // on and double size
	    Sprite[myframe].AffineParamNo_L = 0; // use affine 0
		
		
		LoadEnemyPalette( VFlashPalette,nxtenemypal);
		Sprite[myframe].Pltt = 2 + PalCache;
	
	}
	else
	{
		Sprite[myframe].VPos = 160;
		SpriteScaleFine(0,1,1);
	}
	if(power_scale > 4)
	{
		//if((GameClock & 1) == 0)
		power_scale--;
		SpriteScale(0,power_scale >> 1,power_scale >> 1);
		if(power_scale <= 4)
		{
			Sprite[myframe].VPos = 161;
			Sprite[myframe].AffineMode = 0;
			remove();
		}
	}
}


void RemoveRainSwitch(void)
{
	u8 me;
	me = Object[current_strat].parameter;
	rainswitchtlx[me] = 0;
}

u8 NextRainSwitch(void)
{
	u8 a;
	for(a = 0; a < 8; a++)
		if(rainswitchtlx[a] == 0)
			return a;
			
	return 0;
}

void initrainswitch(void)
{
	u8 me;
	me = NextRainSwitch();
	rainswitchtlx[me] = parm_xpos;
	rainswitchtly[me] = wp_ypos;
	rainswitchbrx[me] = wp_xpos;
	rainswitchbry[me] = parm_ypos;
	Object[num_strats].handle = HandleRainSwitch;
	Object[num_strats].remover = RemoveRainSwitch;
	Object[num_strats].parameter = me;
	itsraining = 0;
	rainactive = 99;
	Screen3(0);
	TextClear();
}

void HandleRainSwitch(void)
{
	u8 me;
	me = Object[current_strat].parameter;
#if 1
	if(player.xpos >= rainswitchtlx[me] && player.xpos < rainswitchbrx[me] && player.ypos >= rainswitchtly[me] && player.ypos < rainswitchbry[me])
	{
	
		if(Game_Area == 3) // ko-wahi
		{
			*(vu16 *)REG_BG3CNT &= (~BG_PRIORITY_0);
			*(vu16 *)REG_BG3CNT |= BG_PRIORITY_2;
		}
		if(itsraining == 0)
		{
			StartRain();
		}
		rainactive = me + 1;
	}
	else
	{
		if(rainactive == me + 1) // just gone out of active area
		{
			if(Game_Area == 3) // ko-wahi
			{
				*(vu16 *)REG_BG3CNT &= (~BG_PRIORITY_2);
				*(vu16 *)REG_BG3CNT |= BG_PRIORITY_0;

				if(itsraining == 0)
				{
					StartRain();
				}
			}
			else
			{
				if(itsraining)
				{
					RemoveRain();
				}
				lightningtime = 0;
			}
			rainactive = -1;
		}
		if(itsraining == 0 && Game_Area == 3)
		{
				*(vu16 *)REG_BG3CNT &= (~BG_PRIORITY_2);
				*(vu16 *)REG_BG3CNT |= BG_PRIORITY_0;

			//	if(itsraining == 0)
			//	{
					StartRain();
			//	}
		}
	}
#endif	
	
	
}

void StartRain(void)
{
	u8 x,y;
	Screen3(0);	
	
	if(Game_Area == 0 || Game_Area == 1)
	{
		DmaArrayCopy(3,raineffectSprite,(u32)(BG_VRAM + 0x3000 + (124 * 32)),32); // last 4 chars
	}
	else
	{
		DmaArrayCopy(3,snoweffectSprite,(u32)(BG_VRAM + 0x3000 + (124 * 32)),32); // last 4 chars
	}
	DmaWait(3);


   *(vu16 *)REG_BG3CNT =                             // Set BG control for BG 1
            BG_COLOR_16 | BG_SCREEN_SIZE_0 | BG_PRIORITY_0
            | 3 << BG_SCREEN_BASE_SHIFT | 0 << BG_CHAR_BASE_SHIFT ;  // font uses chars 512 - 1023


	for(y = 0; y < 32; y++)
		for(x = 0; x < 32 ; x++)
			if(Game_Area != 1)
				*(vu16 *)(BG_VRAM + (2048 * 3) + x * 2 + y * 64) = (min(3,rndfast(7)) + 508) | 0xf000;
			else
				*(vu16 *)(BG_VRAM + (2048 * 3) + x * 2 + y * 64) = (rndfast(3)+ 508) | 0xf000;
				
//	rainontime = 15;
	Screen3(1);
	//BlendAlpha(5,8);
	//*(vu16 *)REG_BLDCNT |= BLD_BG3_1ST;

	if(itsraining == 0)
		spawn(initrainer,0,0,0);
	itsraining = 1;
	if(Game_Area == 0)
	{
		if(Game_Level != 2 || daytime == 352)
			lightningtime = 20 * 60;
	}
	else
	{
		lightningtime = 0;
	}
	//TextColour(31,31,31);
}

void RemoveRain(void)
{
	Screen3(0);
	itsraining = 0;
}

void initrainer(void)
{
	u32 i;
	u8 me;
	u16 x;
	
	Object[num_strats].handle = HandleRainer;
	Object[num_strats].remover = RemoveRain;
	
	rain_x = 0;
	rain_y = 0;		
	raindepth = 0;
	*(vu16 *)BG_PLTT = Lightning[0];

}

void HandleRainer(void)
{
	*(vu16 *)REG_BG3VOFS = rain_y;
	*(vu16 *)REG_BG3HOFS = rain_x;
	
	if(rainontime)
	{
		rainontime--;
		//if(rainontime == 0)
		///	Screen3(1);
	}

	if(Game_Area == 0)
	{
		rain_x += 2;
		rain_y -= 2;
		if(Game_Level != 2) // muaka boss doesn't like discoloring :)
			*(vu16 *)(BG_PLTT + 510) = RainColour[raindepth >> 7]; 
		if(raindepth < 1024)
			raindepth++;		
	}
	else
	{
		if(Game_Area == 1) // duststorm
		{
			rain_x += 5;
			rain_y -= 1;
			*(vu16 *)(BG_PLTT + 510) = 0x4294;
		//	TextColour(20,20,15);
		}
		else
		{
			if(GameClock & 1)
				rain_y --;
			*(vu16 *)(BG_PLTT + 510) = 0x7fff;
		}
	}
}


void StartTorch(void)
{
	u8 x,y;
	

	TorchLitO = 1 - TorchLit;

   *(vu16 *)REG_BG3CNT =                             // Set BG control for BG 1
            BG_COLOR_16 | BG_SCREEN_SIZE_0 | BG_PRIORITY_0
            | 3 << BG_SCREEN_BASE_SHIFT | 0 << BG_CHAR_BASE_SHIFT;  // font uses chars 512 - 1023

	Screen3(1);

	if(TorchLit)
	{
		DmaArrayCopy(3,TorchCoverChar,(u32)(BG_VRAM + 0x3000),32); // overwrites font chars
		DmaWait(3);
		DmaArrayCopy(3,TorchCoverMap,BG_VRAM + 2048 * 3,16);
		DmaWait(3);
		DmaCopy(3,TorchCoverPalette + 480,BG_PLTT + 480,32,16);
		DmaWait(3);
	}
	else
	{
		DmaArrayCopy(3,(u32)DarkCoverChar,(u32)(BG_VRAM + 0x3000),16); // overwrites font chars
		DmaWait(3);
		DmaArrayCopy(3,(u32)DarkCoverMap,BG_VRAM + 2048 * 3,16);
		DmaWait(3);
			DmaCopy(3,TorchCoverPalette + 480,BG_PLTT + 480,32,16);
			DmaWait(3);
	}
	WaterEffect(4,7);
	*(vu16 *)REG_BLDCNT |= BLD_BG3_1ST;
	if(torchactive == 0)
		spawn(inittorch,0,0,0);
	torchactive = 1;
	Screen3(1);
}

void RemoveTorch(void)
{
	Screen3(0);
	torchactive = 0;
	TorchLit = 0;
}

void inittorch(void)
{

	Object[num_strats].handle = HandleTorch;
	Object[num_strats].remover = RemoveTorch;
}

void HandleTorch(void)
{
	u16 * dest;
	

	WaterEffect(4,7);
	*(vu16 *)REG_BLDCNT |= BLD_BG3_1ST;
	if(TorchLit)
	{
		torchcentre = 88;
		torchtop = 16;
		if(TorchLitO == 0)
		{
			DmaArrayCopy(3,TorchCoverChar,(u32)(BG_VRAM + 0x3000),32); // overwrites font chars
			DmaWait(3);
			DmaArrayCopy(3,TorchCoverMap,BG_VRAM + 2048 * 3,16);
			DmaWait(3);
			DmaCopy(3,TorchCoverPalette + 480,BG_PLTT + 480,32,16);
			DmaWait(3);
			TorchLitO = 1; 
			WaterEffect(4,7);
			*(vu16 *)REG_BLDCNT |= BLD_BG3_1ST;
		}
	}
	else
	{
		if(TorchLitO == 1)
		{
			DmaArrayCopy(3,DarkCoverChar,(u32)(BG_VRAM + 0x3000),32); // overwrites font chars
			DmaWait(3);
			DmaArrayCopy(3,DarkCoverMap,BG_VRAM + 2048 * 3,16);
			DmaWait(3);
			DmaCopy(3,TorchCoverPalette + 480,BG_PLTT + 480,32,16);
			DmaWait(3);
			TorchLitO = 0;
			WaterEffect(4,7);
			*(vu16 *)REG_BLDCNT |= BLD_BG3_1ST;
		}
		
	}

}


u8 NextBaddyBlast(void)
{
	u8 a;
	for(a = 0; a < 16; a++)
		if(Blast[a].xpos == 0)
			return a;
	return 0;
}

u8 CheckBlasts(void)
{
	u8 a;
	for(a = 0; a < 16 ; a++)
		if(Blast[a].xpos == 0)
			return 1;
	return 0;
}

void RemoveBlast(void)
{
	u8 me;
	me = Object[current_strat].parameter;
	Blast[me].xpos = 0;
}



void initbaddyblast(void)
{
	u8 frame,me;
	me = NextBaddyBlast();
	
	frame = GrabASprite();
	Blast[me].frame = frame;
	
	Blast[me].clock = 0;
	Blast[me].xpos = parm_xpos;
	Blast[me].ypos = parm_ypos;
	Blast[me].spin = min(2,Object[num_strats].parameter);
	
	Object[num_strats].parameter = me;
	Object[num_strats].handle = HandleBlast;
	Object[num_strats].remover = RemoveBlast;
	if(Blast[me].spin == 0)
	{
		Sprite[frame].ObjMode = 0; // window
    	Sprite[frame].AffineMode = 0; // off 3; // on and double size
    Sprite[frame].AffineParamNo_L = 0; // use affine 0
   	Sprite[frame].Shape = 0;
   	Sprite[frame].Size = 3;
    }
    else
    {
    	Sprite[frame].ObjMode = 0;
    	Sprite[frame].AffineMode = 3;
    Sprite[frame].AffineParamNo_L = 3 + me; // use affine 0
    	Blast[me].dx = srnd(8) * 4;
    	Blast[me].dy = (-rnd(2) - 6) * 4;
   	Sprite[frame].Shape = 0;
   	Sprite[frame].Size = 1; // 16x16
    }
   	Sprite[frame].CharNo = ENEMY_SPRITE;
    Sprite[frame].Priority = 1;
   	Sprite[frame].Pltt = 7;
   	
	
}

void HandleBlast(void)
{
	u8 me,myframe;
	s16 sprx,spry;
	u32 sprite_ptr;
	
	me = Object[current_strat].parameter;

	if(Blast[me].spin)
	{
		Blast[me].xpos += (Blast[me].dx / 4);
		Blast[me].ypos += ( Blast[me].dy / 4);
		sprx = Blast[me].xpos - 16 - game_x;
		spry = Blast[me].ypos - 16 - game_y;
	}
	else
	{
		sprx = Blast[me].xpos - 32 - game_x;
		spry = Blast[me].ypos - 32 - game_y;
	}
	myframe = Blast[me].frame;
	if(sprx > -64 && sprx < 240 && spry > -64 && spry < 160)
	{
		Sprite[myframe].HPos = sprx & 511;
		Sprite[myframe].VPos = (spry + 256) & 255;
		
		if(Blast[me].spin == 0)
		{
			sprite_ptr = (u32)BaddyDeathSprite + (Blast[me].clock >> 2 ) * 2048;
			LoadEnemySprite(sprite_ptr, 2048);
			LoadEnemyPalette(BaddyDeathPalette,nxtenemypal);
			Sprite[myframe].AffineMode = 0;
			Sprite[myframe].Pltt = 2 + PalCache;
		}
		else
		{
			sprite_ptr = (u32)Particle2[Game_Area] + (u32)(128 + 128 * Blast[me].spin);
			LoadEnemySprite(sprite_ptr, 128);
			Sprite[myframe].Pltt = 15;

		//	LoadEnemyPalette(ParticleP[Game_Area],nxtenemypal);
			SpriteRotateScale(3 + (me >> 1),((GameClock & 63) * 2 * Blast[me].spin) & 255,min(8,1 + (Blast[me].clock >> Blast[me].spin)),min(8,1 + (Blast[me].clock >> Blast[me].spin)));
		   	Sprite[myframe].HFlip = 0;
		   	Sprite[myframe].VFlip = 0;
		   	Sprite[myframe].AffineMode = 3;
		    Sprite[myframe].AffineParamNo_L = 3 + (me >> 1); // use affine 0
			
		}
		Sprite[myframe].CharNo = EnemySprite[SpriteCache];
	}
	else
	{
		Sprite[myframe].VPos = 160;
		if(Blast[me].spin)
		{
			Sprite[myframe].VPos = 161;
			Sprite[myframe].AffineMode = 0;
			remove();
			return;
		}
	}
	Blast[me].clock++;
	if(Blast[me].spin == 0)
	{
		if(Blast[me].clock > 15)
		{

			Sprite[myframe].VPos = 161;
			Sprite[myframe].AffineMode = 0;
			SpawnParticleBlast(Blast[me].xpos,Blast[me].ypos);
			remove();
			return;

		}
	}
	else
	{
		Blast[me].dy++;
		if(ForceOtherCollision(Blast[me].xpos,Blast[me].ypos) < Blast[me].ypos) // bounce
		{
			Blast[me].dy = -abs(Blast[me].dy) + 8;
			if(Blast[me].dy > -4 && Blast[me].clock > 30)
			{
				Sprite[myframe].VPos = 161;
				Sprite[myframe].AffineMode = 0;
				remove();
			}
			
		}
		
	}

}



u8 inittrailsprite(s16 xpos,s16 ypos,u16 CharNo,u8 size,u8 shape,u8 pal)
{
	u8 frame,me,a;
	
	if(nxttrail < 16)
	{
		me = nxttrail++;
	//	frame = GrabASprite();

		trail[me].charno = CharNo;
		trail[me].xpos = xpos;
		trail[me].ypos = ypos;
		trail[me].time = 8;
		trail[me].frame = frame;
		trail[me].size = size;
		trail[me].shape = shape;
		trail[me].pal = pal;
	}
}

void handletrailsprites(void)
{
	s8 a,b;
	s16 sprx,spry;
	u8 frame;
	
	for(a = 0; a < nxttrail;a++)
	{
		trail[a].time--;
		sprx = trail[a].xpos - game_x;
		spry = trail[a].ypos - game_y;
	//	frame = trail[a].frame;
		if(sprx > -64 && sprx < 240 && spry > -64 && spry < 160 && trail[a].time > 0)
		{
			if((GameClock & 1) == (a & 1))
			{
				frame = NextSprite++;
				
				Sprite[frame].ObjMode = 0; // trans effect
			    Sprite[frame].AffineMode = 0; // off 
			    Sprite[frame].AffineParamNo_L = 0; // use affine 0
			   	Sprite[frame].CharNo = trail[a].charno;
			   	Sprite[frame].Shape = trail[a].shape;
			   	Sprite[frame].Size = trail[a].size;
			    Sprite[frame].Priority = 2;
			   	Sprite[frame].Pltt = trail[a].pal;
				Sprite[frame].HPos = sprx & 511;
				Sprite[frame].VPos = (spry + 256) & 255;
			}
		}		
		else
		{
			trail[a].time = 0;
			b = nxttrail - 1;
			if(a < b)
				trail[a] = trail[b];
			nxttrail--;
			a--;
			
		}
		
	}
}

void SetSpriteColour(u8 index,u8 red, u8 green, u8 blue)
{
	u16 bgr;
	u16 * ptr;
	
	bgr = red + (green << 5) + (blue << 10);
	ptr = OBJ_PLTT + index * 2;
	*ptr = bgr;
}

