//#define CODE32
#include <agb.h>
#include "share.h"
#include "menu.h"
#include "menuitem.h"
#include "SampleManager.h"
#include "Display.h"
#include "FontPrn.h"
#include "EngineTest.h"
#include "Thala.h"
#include "Polys.h"
#include "DrawPoly.h"
#include "Pool.h"

#define VCOUNT  (*(volatile unsigned short*)0x04000006)

static void my_v_sync_fnc2(void);

#if 1
volatile int PALETTE;
#else
#define PALETTE (*(volatile unsigned short*)0x05000000)
#endif

#define min(a, b)  (((a) < (b)) ? (a) : (b)) 
#define max(a, b)  (((a) > (b)) ? (a) : (b)) 

int lastrandom = 0;
int random(void)
{
	if (lastrandom == 0)
	{
		lastrandom = 1234567;
	}

	lastrandom *= 7654321;
	lastrandom += 378271;
	return (lastrandom&65535)^((lastrandom>>16)&65535);
}

void EngineTest_Entry(void)
{
	int i;
	Polys polys;

	*(vu16 *)REG_WAITCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT
						 | CST_PREFETCH_ENABLE;			// Set up 3-1 wait access
	//  Clear RAM
	ClearRamAll();
	gPool.Init((void *)EX_WRAM,EX_WRAM_SIZE);
//	gPool.Init((void *)CPU_WRAM,CPU_WRAM_SIZE);

	SampleManager_Init(1);
	gDisplay.Init();
	polys.Init(300);

	for (i=0;i<256;i++)
	{
		((u16 *)PLTT)[i] = (i&15) | ((i&31)<<5) | ((i&15)<<11);
	}

	//  Initialize V Blank 

	InitVBlank();

	*(vu16 *)REG_IE   = 0;//CASSETTE_INTR_FLAG;	 // Enable Game Pak Interrupt
	*(vu16 *)REG_IME  = 1;					  // Set IME 
	*(vu16 *)REG_STAT = STAT_V_BLANK_IF_ENABLE;	// V Blank interrupt request flag enabled 

	//  For debugging
	KeyInit();
	KeyRead();

	//  Set V Blank function 

	SetVBlankFunc( my_v_sync_fnc2 );

	*(vu16 *)REG_IE |= V_BLANK_INTR_FLAG;

	Mod_Init(gThalamusPatternList,gThalamusSamples);
	(*(volatile unsigned short*)0x05000000) = 0xff;

	i = 64;
	int cycle = 0;
#if 1
		polys.Clear();

#define NUMPOLYS (20)

		for (int j=0;j<NUMPOLYS;j++)
		{
			ScreenPoly *poly;

			poly = polys.GetNewPoly();
			poly->mType = ScreenPoly::kFlat;
			poly->mColours[0] = ((j&15)+1);

#define POLYSIZEX (240)
#define POLYSIZEY (160)

			poly->mX[0] = random()%POLYSIZEX;	poly->mY[0] = random()%POLYSIZEY;
			poly->mX[1] = random()%POLYSIZEX;	poly->mY[1] = random()%POLYSIZEY;
			poly->mX[2] = random()%POLYSIZEX;	poly->mY[2] = random()%POLYSIZEY;
			polys.SubmitPoly(poly,random()&65535);
		}
#endif
	while(1)
	{
		void *screenmem;

		gDisplay.Swap();
		PALETTE = 0x00;
		V_Wait();
		PALETTE = 0xff;

		gDisplay.Clear();
		screenmem = gDisplay.GetBackScreen();

		polys.Clear();
		for (int j=0;j<NUMPOLYS;j++)
		{
			ScreenPoly *poly;

			poly = polys.GetNewPoly();
			poly->mX[0]++;	if (poly->mX[0] >= POLYSIZEX) poly->mX[0] -= POLYSIZEX;	if (poly->mX[0] < 0) poly->mX[0] += POLYSIZEX;
			poly->mX[1]--;	if (poly->mX[1] >= POLYSIZEX) poly->mX[1] -= POLYSIZEX;	if (poly->mX[1] < 0) poly->mX[1] += POLYSIZEX;
			poly->mX[2]++;	if (poly->mX[2] >= POLYSIZEX) poly->mX[2] -= POLYSIZEX;	if (poly->mX[2] < 0) poly->mX[2] += POLYSIZEX;
			poly->mY[0]--;	if (poly->mY[0] >= POLYSIZEY) poly->mY[0] -= POLYSIZEY;	if (poly->mY[0] < 0) poly->mY[0] += POLYSIZEY;
			poly->mY[1]--;	if (poly->mY[1] >= POLYSIZEY) poly->mY[1] -= POLYSIZEY;	if (poly->mY[1] < 0) poly->mY[1] += POLYSIZEY;
			poly->mY[2]++;	if (poly->mY[2] >= POLYSIZEY) poly->mY[2] -= POLYSIZEY;	if (poly->mY[2] < 0) poly->mY[2] += POLYSIZEY;
			polys.SubmitPoly(poly,random()&65535);
		}
//		polys.Sort();
		polys.Draw(gDisplay.GetBackScreen());

		KeyRead();

		if( Cont & A_BUTTON )
		{
			i++;
			if (i>=160)
			{
				i = 0;
			}
		}

		PALETTE = 0xff0;
	}
}

static void my_v_sync_fnc2(void)
{
	SampleManager_VSyncCode();
	gDisplay.VSyncFunction();
	SampleManager_OnceAFrame();
}
