#ifndef __VECTORMATHS_H__
#define __VECTORMATHS_H__

/* Some defines */

typedef int SCALAR;

typedef SCALAR t_vec;
typedef t_vec t_vec2[2];
typedef t_vec t_vec3[3];

/* Some vector maths operations */
#define Vector_Sub(dest,src1,src2)	( (dest)[0] = (src1)[0] - (src2)[0]);\
					( (dest)[1] = (src1)[1] - (src2)[1]);\
					( (dest)[2] = (src1)[2] - (src2)[2]);

#define Vector_Add(dest,src1,src2)	( (dest)[0] = (src1)[0] + (src2)[0]);\
					( (dest)[1] = (src1)[1] + (src2)[1]);\
					( (dest)[2] = (src1)[2] + (src2)[2]);

#define Vector_Copy(dest,src)	(dest)[0] = (src)[0];\
				(dest)[1] = (src)[1];\
				(dest)[2] = (src)[2];

/* Conversion from int&float to scalar and back again */
#define ScalarToInt(x)  (int) ((x)>>16)
#define IntToScalar(x)  (SCALAR) ((x)<<16)

#define ScalarToFloat(x)  (float) ((double)(x)/(double)(1<<16))
#define FloatToScalar(x)  (SCALAR) ((float)(x)*(double)(1<<16))

/* Scalar adding, multiply and divide */

#define ScalarAdd(x,y) (SCALAR) ((x)+(y))
#define Scalar3Add(x,y,z) (SCALAR) ((x)+(y)+(z))
#define ScalarSub(x,y) (SCALAR) ((x)-(y))
#define Scalar3Sub(x,y,z) (SCALAR) ((x)-(y)-(z))
#define ScalarMul(x,y) (SCALAR) (((x)>>8)*((y)>>8))
#define Scalar3Mul(x,y,z) (SCALAR) (((x)>>8)*((y)>>8)*((z)>>8))
#define ScalarDiv(x,y) (SCALAR) (((x)<<8)/((y)>>8))
#define ScalarDiv2(x,y) (SCALAR) (((x)<<8)/(y))
#define ScalarSqrt(x) (SCALAR) ( FloatToScalar(sqrt(ScalarToFloat(x))) )
#define ScalarMulAcc(x,y) (SCALAR) ( FloatToScalar(ScalarToFloat(x) * ScalarToFloat(y)) )
#define BR_MAC3(a,b,c,d,e,f) Scalar3Add(asm_fx_AccurateMul(a,b),asm_fx_AccurateMul(c,d),asm_fx_AccurateMul(e,f))

extern SCALAR asm_fx_AccurateMul(SCALAR,SCALAR);

typedef struct {
  t_Plotter      Draw;
  t_Plotter      DrawTextured;
  t_Plotter      DrawTexturedTrans;
  t_Plotter      DrawGouraud;
  t_Plotter      DrawTexturedGouraud;
  t_Plotter      DrawTexturedGouraudTrans;
  t_Plotter      DrawAlpha;
  t_Plotter      DrawTexturedAlpha;
  t_Plotter      DrawTexturedAlphaTrans;
  t_Plotter      DrawAlphaGouraud;
  t_Plotter      DrawTexturedAlphaGouraud;
  t_Plotter      DrawTexturedAlphaGouraudTrans;
  t_Plotter      DrawUC;
  t_Plotter      DrawTexturedUC;
  t_Plotter      DrawTexturedTransUC;
  t_Plotter      DrawGouraudUC;
  t_Plotter      DrawTexturedGouraudUC;
  t_Plotter      DrawTexturedGouraudTransUC;
  t_Plotter      DrawAlphaUC;
  t_Plotter      DrawTexturedAlphaUC;
  t_Plotter      DrawTexturedAlphaTransUC;
  t_Plotter      DrawAlphaGouraudUC;
  t_Plotter      DrawTexturedAlphaGouraudUC;
  t_Plotter      DrawTexturedAlphaGouraudTransUC;
} PlotClass;

extern PlotClass Plotters_8bit;
extern PlotClass Plotters_16bit;
extern PlotClass Plotters_24bit;

void PlotLib_SetFrameBufferSize(int width, int height);

void PlotLib_SetFrameBufferStart(char *memory);

#endif
