/******************************************/
/*										  */
/*  ASL Defines for GBA					  */
/* Aaron Fothergill, Argonaut Software    */
/* Oct 2001								  */
/******************************************/
#ifndef _ASL_DEFS_H
#define _ASL_DEFS_H

#include <AGB.h>

#define DEBUG_MODE 1

#define DEBUG_OUT(x,y) AGBPrintf(x,y); \
	AGBPrintFlush();  

#define LoadSprite(x,y,z) DmaCopy(3,(u32)x,(u32)(OBJ_MODE0_VRAM + y),z,16); DmaWait(3)
#define LoadSpriteFont(x) DmaArrayCopy(3,(u32)x,(u32)(BG_VRAM + 0xe000),16); DmaWait(3)
#define LoadPalette(x,y) DmaCopy(3,(u32)x,(u32)(OBJ_PLTT + y * 32),32,16); DmaWait(3)
#define LoadTiles(x,y) DmaArrayCopy(3,(u32)x,(u32)(BG_VRAM + y),16); DmaWait(3)
#define LoadMapPalette(x) DmaArrayCopy(3,x,BG_PLTT,16); CurrentPalette = (u32)x ; DmaWait(3)
#define RestoreMapPalette(x) DmaCopy(3,x,BG_PLTT + 32,480,16); DmaWait(3)
#define LoadMapPaletteb(x) DmaArrayCopy(3,x,BG_PLTT,16); DmaWait(3);
#define LoadMapPalette2(x,y) DmaCopy(3,x,(BG_PLTT + y * 32),32,16); DmaWait(3)
#define BlendAlpha(x,y)  *(vu16 *)REG_BLDCNT =  BLD_A_BLEND_MODE |  BLD_2ND_ALL; *(vu16 *)REG_BLDALPHA = (x) + ((y) << 8)
#define BlendDown(x) 	*(vu16 *)REG_BLDCNT =  BLD_DOWN_MODE |  BLD_1ST_ALL; *(vu16 *)REG_BLDY = (x)
#define BlendUp(x)  	*(vu16 *)REG_BLDCNT =  BLD_UP_MODE |  BLD_1ST_ALL;  *(vu16 *)REG_BLDY = (x);
#define BlendDownScreens(x) 	*(vu16 *)REG_BLDCNT =  BLD_DOWN_MODE |  BLD_BG0_1ST | BLD_BG1_1ST | BLD_BG2_1ST | BLD_BG3_1ST; *(vu16 *)REG_BLDY = (x)
#define BlendDownPause(x)		*(vu16 *)REG_BLDCNT = BLD_DOWN_MODE | BLD_BG0_1ST | BLD_BG1_1ST | BLD_BG2_1ST | BLD_OBJ_1ST; *(vu16 *)REG_BLDY = (x)
#define WaterEffect(x,y) *(vu16 *)REG_BLDCNT =  BLD_A_BLEND_MODE | BLD_BG1_1ST | BLD_OBJ_2ND | BLD_2ND_ALL; *(vu16 *)REG_BLDALPHA = (x) + ((y) << 8);
#define LoadEnemySprite(x,z)  CheckSpriteCache(x,z) 
#define LoadEnemyPalette(x,y)  { CheckPalCache(); DmaCopy(3,(u32)x,(u32)(OBJ_PLTT + EnemyPaletteOffset[PalCache]),32,16); DmaWait(3) }
#define LoadEnemyCyclePalette(x,y) DmaCopy(3,(u32)x,(u32)(OBJ_PLTT + EnemyPaletteOffset[PalCache] + 26),6,16); DmaWait(3)
#define LoadEnemyCycle2Palette(x,y) DmaCopy(3,(u32)x,(u32)(OBJ_PLTT + EnemyPaletteOffset[PalCache] + 30),6,16); DmaWait(3)
#define LoadPlayerCyclePalette(x,y) *(vu16 *)(OBJ_PLTT + (y * 32) + 14) = x;
#define LoadChiefCyclePalette(x,y) *(vu16 *)(OBJ_PLTT + (y * 32) + 28) = x;
// Controller defines...

#define controller_read		(*(vu16 *)REG_KEYINPUT ^ 0x3ff)
#define pad_left		L_KEY
#define pad_right		R_KEY
#define pad_forward		U_KEY
#define pad_backward	D_KEY
#define pad_up			U_KEY
#define pad_down		D_KEY
#define pad_start		START_BUTTON
#define pad_select		SELECT_BUTTON
#define pad_all_dir		PLUS_KEY_MASK
#define pad_all_buttons	BUTTON_MASK
#define pad_D			(L_KEY | R_KEY | U_KEY | D_KEY)

u16 pad_A;
u16 pad_B;
u16 pad_L;
u16 pad_R;

#define MAX_STRATS		140
const s16 fsin[256];
typedef void (*Funcp)(void) ;


typedef struct {
	Funcp handle;		// handler routine
	Funcp remover;		// strat removal routine
	s32	parameter;
	u8 stratnum;		// number in strat list (255 = not in list)
	u8 respawnme;		// set to 1 to cause respawn on player death
} object_type;


// controller globals

u16 controller, controller_push;
u16 current_strat,num_strats;
object_type Object[MAX_STRATS];
OamData  Sprite[128];                // OAM backup
s32 parm_xpos,parm_ypos;
s32 wp_xpos,wp_ypos;

u16	collision_underme;		// nearest ground point beneath player
u8 collision_left;			// floor of left collision	
u8 collision_right;		// floor of right collision
u8 collision_left_ladder;
u8 collision_right_ladder;
u16 collision_floornum;		// floor number of last collision returned
u16 collision_aboveme;   // nearest ceiling point above player
u16 platform_underme;		// nearest platform beneath player
u16 platform_underme2;
u16 Floor_Depth;

u16 * col_map;
u8 * col_tile_fine;
u32 CurrentPalette;

u8 NextSprite;		// next sprite to be used
u8 LastSprite;      // Last sprite used in previous frame
u8 mysprite;


void KeyRead(void);
void NoCollision(void);
void spawn(Funcp stratinit,s16 xpos,s16 ypos,s32 parameter);
void handle_strats(void);
void delete_strat(void);
void respawnme(void);
void remove(void); 
void removeall(void);

u8 GrabASprite(void);
s16 sgn(s16 val);
void ForceCollision(u16 object_xpos,u16 object_ypos);
s16 ForceOtherCollision(u16 object_xpos,u16 object_ypos);
void WallCollision(u16 object_xpos,u16 object_ypos);
void CeilingCollision(u16 object_xpos,u16 object_ypos);

s32 min(s32 a,s32 b);
s32 max(s32 a,s32 b);
u32 abs(s32 a);
u32 rnd(u32 a);
s32 srnd(u32 a);
u32 rndfast(u32 a);
void UpdateRandomClock(void);
void SpriteRotate(u8 myaffine , u8 ang);
void SpriteScale(u8 myaffine, s8 scalex,s8 scaley);
void SpriteScaleFine(u8 myaffine, s16 scalex,s16 scaley);
void SpriteRotateScale(u8 myaffine, u8 ang, s8 scalex,s8 scaley);
void CycleColour(u16 palette,u16 start,u16 length);
void CycleMapColour(u16 palette,u16 start,u16 length);
void CycleMapColourB(u16 palette,u16 start,u16 length);
void PasteSprite(s16 xpos,s16 ypos,u8 shape,u8 size,u16 charno);
void CheckSpriteCache(u32 x,u32 z);
u8 CheckPalCache(void);
void SetMapColour(u8 index,u8 red, u8 green, u8 blue);
void FlashMapColour(u8 index,u8 red,u8 green,u8 blue);
void DimMapPalette(u32 pal);
void SaveGameParameters(void);
void LoadGameParameters(void);
void LoadCharAvail(void);
void SaveCharAvail(void);
void FlushSpriteCache(void);


#endif
